@@ -4,7 +4,7 @@ use warnings FATAL => 'all';
 
 package namespace::local;
 
-our $VERSION = '0.07';
+our $VERSION = '0.08';
 
 =head1 NAME
 
@@ -13,9 +13,36 @@ namespace::local - Confine imports or functions to a given scope
 =head1 SYNOPSIS
 
 This module allows to confine imports or private functions
-to a given scope. The following modes of operation exist:
+to a given scope.
 
-=head2 -around (the default)
+    package My::Module;
+
+    sub normal_method {
+        # frobnicate; # nope!
+    };
+
+    sub method_with_sugar {
+        use namespace::local;
+        use Crazy::Prototyped::DSL qw(do_this do_that frobnicate);
+
+        do_this;
+        do_that;
+        frobnicate;
+    };
+
+    sub another_method {
+        # frobnicate; # nope!
+    };
+
+The calling module's symbol table is saved at the C<use> line
+and restored upon leaving the block.
+
+The subsequent imports will do their job within the block,
+but will not be available as methods at runtime.
+
+=head1 MODES OF OPERATION
+
+=head2 -around
 
 This confines all subsequent imports and functions
 between the use of L<namespace::local> and the end of scope.
@@ -27,7 +54,7 @@ between the use of L<namespace::local> and the end of scope.
     }
 
     sub using_import {
-        use namespace::local;
+        use namespace::local -around;
         use Some::Crazy::DSL qw(frobnicate);
         frobnicate Foo => 42;
     }
@@ -36,7 +63,7 @@ between the use of L<namespace::local> and the end of scope.
         # frobnicate() is unknown
     }
 
-=head2 -below
+=head2 -below (the default)
 
 Hides subsequent imports and functions on end of scope.
 
@@ -98,6 +125,18 @@ from the module's action.
 
 Note that sigils are ignored here.
 
+=head2 -only => \@list
+
+Only affect the listed symbols (with sigils).
+Rules are the same as for -except.
+
+=head2 -only => <regex>
+
+Only affect symbols with matching names.
+
+All C<-only> and C<-except> options act together, further restricting the
+set of affected symbols.
+
 =head1 EXEMPTIONS
 
 The following symbols are not touched by this module, to avoid breaking things:
@@ -110,17 +149,19 @@ The following symbols are not touched by this module, to avoid breaking things:
 
 =item * Arrays: C<@CARP_NOT>, C<@EXPORT>, C<@EXPORT_OK>, C<@ISA>;
 
-=item * Scalars: C<$AUTOLOAD>, C<$a>, C<$b>;
+=item * Hashes: C<%OVERLOAD>;
 
 =item * Files: C<DATA>, C<STDERR>, C<STDIN>, C<STDOUT>;
 
 =item * Functions: C<AUTOLOAD>, C<DESTROY>, C<import>;
 
+=item * Scalars: C<$AUTOLOAD>, C<$a>, C<$b>;
+
 =back
 
 This list is likely incomplete, and may grow in the future.
 
-=head1 METHOD/FUNCTIONS
+=head1 METHODS/FUNCTIONS
 
 None.
 
@@ -151,12 +192,15 @@ Due to order of callback execution in L<B::Hooks::EndOfScope>,
 other modules in C<namespace::> namespace may interact poorly
 with L<namespace::local>.
 
-However, at least C<-above> and C<-below> switches
-work as expected if used simultaneously.
+Up to v.0.07, C<-around> used to be the default mode instead of C<-below>.
+C<-around> is much more restrictive, in particular, it prevents functions
+defined below the block from propagating above the block.
+
+This is less of a problem than imported functions leaking upward.
+No perfect solution has yet been found.
 
 =cut
 
-# this was stolen from namespace::clean
 use B::Hooks::EndOfScope 'on_scope_end';
 
 my @stack;
@@ -167,7 +211,7 @@ sub import {
     my $command = namespace::local::_command->new( caller => [ caller ] );
     $command->parse_options( @_ );
 
-    # on_scope_end callback execution order is direct
+    # on_scope_end executes multiple callbacks as FIFO
     # we need reversed order, so use a stack of commands.
     $stack[-1]->set_next( $command ) if @stack;
     push @stack, $command;
@@ -175,35 +219,89 @@ sub import {
     $command->prepare;
 
     on_scope_end {
+        local $Carp::Internal{'B::Hooks::EndOfScope::XS'} = 1;
+        local $Carp::Internal{'B::Hooks::EndOfScope'} = 1;
         pop @stack;
         $command->execute;
     };
 };
 
-# Hide internal OO engine
-# Maybe it will be released later...
+=head1 ENVIRONMENT
+
+Set C<PERL_NAMESPACE_LOCAL=debug> to see some debugging information
+upon module load.
+
+=head1 THE INTERNALS
+
+A stack of "command" objects is used behind the scenes.
+
+Its interface is not public, see this module's source.
+
+Calling
+
+=over
+
+=item new
+
+=back
+
+on this package will create a command object,
+I<not> a C<namespace::local> instance.
+
+The creation and destruction of command has no effect on the namespaces.
+
+Instead, special C<prepare> and C<execute> methods
+are called upon import and leaving scope, respectively.
+
+=cut
+
+sub new {
+    my $unused = shift;
+    namespace::local::_command->new(caller => [caller 0], @_);
+};
 
 package
     namespace::local::_command;
 
-use Carp; # too
+# FIRST AND FOREMOST
+
+# See C<perldoc perlmod> for how perl stores the Symbol Tables.
+
+# In this module we use two-level hashrefs to represent the table:
+# $table->{ $name }{ $type } = $reference
+# where $name is a function/variable/whatever name,
+# $type is one of ARRAY, CODE, FORMAT, HASH, IO, and SCALAR (see @TYPES below),
+# and $reference is a reference of corresponding type (or undef).
+
+# So Foo::bar() would be represented as $table->{foo}{CODE}
+# whereas @Foo::ISA is $table->{ISA}{ARRAY}.
+
+use Carp;
+use Scalar::Util qw(blessed refaddr reftype);
 our @CARP_NOT = qw(namespace::local);
 
-# target => package_name is required
+# TODO need better env parsing...
+use constant DEBUG => ( lc ($ENV{PERL_NAMESPACE_LOCAL} || '' ) eq 'debug' ? 1 : 0 );
+
+### Setup methods
+
 sub new {
     my ($class, %opt) = @_;
 
     # TODO check options
-    $opt{except_rex} = qr/^[0-9]+$|^_$/; # no matter what, exempt $_, $1, ...
-    $opt{action} = '-around';
-    $opt{target} = $opt{caller}[0];
-    $opt{origin} = join ":", @{$opt{caller}}[1,2];
+    $opt{caller}     ||= [ caller 0 ];
+    $opt{except_rex} ||= qr/^[0-9]+$|^_$/; # no matter what, exempt $_, $1, ...
+    $opt{only_rex}   ||= qr/^/; # match all
+    $opt{action}     ||= '-below';
+    $opt{target}     ||= $opt{caller}[0];
+    $opt{origin}     ||= join ":", @{$opt{caller}}[1,2];
 
     # Skip some well-known variables and functions
     # Format: touch_not{ $name }{ $type }
     # NOTE if you change the list, also change the EXEMPTIONS section in the POD.
     $opt{touch_not}{$_}{ARRAY}++  for qw( CARP_NOT EXPORT EXPORT_OK ISA );
     $opt{touch_not}{$_}{CODE}++   for qw( AUTOLOAD DESTROY import );
+    $opt{touch_not}{$_}{HASH}++   for qw( OVERLOAD );
     $opt{touch_not}{$_}{IO}++     for qw( DATA STDERR STDIN STDOUT );
     $opt{touch_not}{$_}{SCALAR}++ for qw( AUTOLOAD a b );
 
@@ -221,17 +319,19 @@ sub set_next {
 
 sub DESTROY {
     my $self = shift;
-    carp "probably a bug in namespace::local: callback set but not executed"
+    carp "probably a bug in namespace::local: callback set at $self->{origin} but never executed"
         if $self->{todo} and !$self->{done};
 };
 
 my %known_action;
 $known_action{$_}++ for qw(-above -below -around);
 
-# this changes nothing except the object itself
+# This changes nothing except the object itself
+# input is the same as that of namespace::local->import
 sub parse_options {
     my $self = shift;
 
+    # wrote a Getopt::Long from scratch...
     while (@_) {
         my $arg = shift;
         if ( $known_action{$arg} ) {
@@ -247,13 +347,38 @@ sub parse_options {
             } else {
                 _croak( "-except argument must be regexp or array" )
             };
+        } elsif ($arg eq '-only') {
+            my $cond = shift;
+            if (ref $cond eq 'Regexp') {
+                $self->{only_rex} = $cond;
+            } elsif (ref $cond eq 'ARRAY') {
+                $self->restrict( @$cond );
+            } else {
+                _croak( "-only argument must be regexp or array" )
+            };
         } else {
             _croak( "unknown option $arg" );
         };
     };
 };
 
-# TODO join with @TYPE array from below
+sub touch_not {
+    my ($self, @list) = @_;
+
+    foreach (sigil_to_type(@list)) {
+        $self->{touch_not}{ $_->[0] }{ $_->[1] }++
+    };
+};
+
+sub restrict {
+    my ($self, @list) = @_;
+
+    foreach (sigil_to_type(@list)) {
+        $self->{restrict_symbols}{ $_->[0] }{ $_->[1] }++
+    };
+};
+
+# TODO join with @TYPES array from below
 my %sigil = (
     ''  => 'CODE',
     '$' => 'SCALAR',
@@ -261,17 +386,20 @@ my %sigil = (
     '@' => 'ARRAY',
 );
 
-sub touch_not {
-    my ($self, @list) = @_;
-
-    foreach (@list) {
+# returns [ name, type ] for each argument
+sub sigil_to_type {
+    map {
         /^([\$\@\%]?)(\w+)$/
             or _croak( "cannot exempt sybmol $_: unsupported format" );
-        $self->{touch_not}{$2}{ $sigil{$1} }++
-    };
+        [ $2, $sigil{$1} ]
+    } @_;
 };
 
-# side effects + setup self->execute
+### Command pattern split into prepare + execute
+
+# input: none
+# output: none
+# side effects: modify target package + setup callback for self->execute
 sub prepare {
     my $self = shift;
 
@@ -279,14 +407,16 @@ sub prepare {
 
     my $table = $self->read_symbols;
 
-    # FIXME UGLY HACK
-    # Immediate backup-and-restore of symbol table
-    #     somehow forces binding of symbols
-    #     above 'use namespace::local' line
-    #     thus preventing subsequent imports from leaking upwards
-    # I do not know why it works, it shouldn't.
     if ($action eq '-around') {
-        $self->replace_symbols( undef, $table );
+        # Overwrite symbol table with a copy of itself.
+        # Somehow this triggers binding of symbols in the code
+        #    that was parsed so far (i.e. above the 'use' line)
+        #    and undefined symbols (in that area) remain so forever
+        $self->write_symbols( $table );
+    } elsif ( $action eq '-below' ) {
+        # Stabilize known functions, leave everything else as is
+        my $func = $self->filter_functions( $table );
+        $self->write_symbols( $func );
     };
 
     if ($action eq '-above' ) {
@@ -300,6 +430,9 @@ sub prepare {
     };
 };
 
+# input: none
+# output: none
+# side effect: modify target package
 sub execute {
     my ($self) = @_;
 
@@ -311,33 +444,13 @@ sub execute {
         unless $self->{done}++;
 };
 
-# Don't touch NAME, PACKAGE, and GLOB itself
+### High-level effectful functions
+
+# Don't ever touch NAME, PACKAGE, and GLOB that are also known to Perl
 my @TYPES = qw(SCALAR ARRAY HASH CODE IO FORMAT);
 
-# in: package name
-# out: sorted & filtered list of symbols
-
-# FIXME needs explanation
-# We really need to filter because copying ALL table
-#     was preventing on_scope_end from execution
-#     (accedental reference count increase?..)
-
-sub read_names {
-    my $self = shift;
-
-    my $package = $self->{target};
-    my $except = $self->{except_rex};
-
-    my @list = sort grep {
-        /^\w+$/ and $_ !~ $except
-    } do {
-        no strict 'refs'; ## no critic
-        keys %{ $package."::" };
-    };
-
-    return @list;
-};
-
+# In: symbol table hashref
+# Out: side effect
 sub erase_only_symbols {
     my ($self, $table) = @_;
 
@@ -357,13 +470,154 @@ sub erase_only_symbols {
     $self->replace_symbols( \@list, $current );
 };
 
-# In: package, symbol
-# Out: a hash with glob content
+# This method's signature is a bit counterintuitive:
+# $self->replace_symbols( \@names_to_erase, \%new_table_entries )
+# If first argument is omitted, the whole namespace is scanned instead.
+# Separate erase_symbols and write_symbols turned out to be cumbersome
+#    because of the need to handle granular exclusion list.
+# This method can fill both roles.
+# Providing an empty list would make it just write the symbol table,
+#    whereas an empty hash would mean deletion only.
+sub replace_symbols {
+    my ($self, $clear_list, $table) = @_;
+
+    $clear_list ||= [ $self->read_names ];
+    $table ||= {};
+
+    my %uniq;
+    $uniq{$_}++ for keys %$table, @$clear_list;
+
+    # re-read the symbol table
+    my $old_table = $self->read_symbols( [ keys %uniq ] );
+
+    # create a plan for change
+    my $diff = $self->table_diff( $old_table, $table );
+    return unless keys %$diff;
+
+    $self->write_symbols( $diff );
+};
+
+sub filter_functions {
+    my ($self, $table) = @_;
+
+    my %new_table;
+
+    foreach (keys %$table) {
+        $new_table{$_} = $table->{$_}
+            if defined $table->{$_}{CODE};
+    };
+
+    return \%new_table;
+};
+
+# Oddly enough, pure
+# In: old and new two symbol table hashrefs
+# Out: part of new table that differs from the old,
+#      with touch_not rules applied
+sub table_diff {
+    my ($self, $old_table, $new_table) = @_;
+
+    my %uniq_name;
+    $uniq_name{$_}++ for keys %$old_table, keys %$new_table;
+
+    my $touch_not = $self->{touch_not};
+
+    if (my $restrict = $self->{restrict_symbols}) {
+        # If a restriction is in place, invert it and merge into skip
+        # TODO write this better
+        # TODO does it really belong here?
+        $restrict->{$_} or delete $uniq_name{$_} for keys %uniq_name;
+        my %real_touch_not;
+        foreach my $name (keys %uniq_name) {
+            # 2 levels of shallow copy is enough
+            foreach my $type( @TYPES ) {
+                $real_touch_not{$name}{$type}++
+                    unless $restrict->{$name}{$type} and not $touch_not->{$name}{$type};
+            };
+        };
+        $touch_not = \%real_touch_not;
+    };
+
+    my $diff;
+
+    # iterate over keys of both, 2 levels deep
+    foreach my $name (sort keys %uniq_name) {
+        my $old  = $old_table->{$name} || {};
+        my $new  = $new_table->{$name} || {};
+        my $skip = $touch_not->{$name} || {};
+
+        my %uniq_type;
+        $uniq_type{$_}++ for keys %$old, keys %$new;
+
+        foreach my $type (sort keys %uniq_type) {
+            next if $skip->{$type};
+
+            if (ref $old->{$type} ne ref $new->{$type}) {
+                # As nonrefs are not allowed here,
+                # this also handles undef vs. defined case
+                $diff->{$name}{$type} = $new->{$type};
+                next;
+            };
+
+            # both undef, nothing to see here
+            next unless ref $new->{$type};
+
+            # pointing to different things
+            if (refaddr $old->{$type} != refaddr $new->{$type}) {
+                $diff->{$name}{$type} = $new->{$type};
+                next;
+            };
+        };
+
+        if ($diff->{$name}) {
+            # if we cannot avoid overwriting,
+            # make sure to copy ALL skipped values we know of
+            $diff->{$name}{$_} = $old->{$_} for keys %$skip;
+
+            # removing a scalar copletely didn't work very well on Perl 5.10.1,
+            # causing segfaults in unrelated places.
+            $diff->{$name}{SCALAR} ||= \undef;
+        };
+    };
+
+    return $diff;
+};
+
+### Low-level symbol-table read & write
+### no magic should happen above this line
+
+# NOTE that even here we are in full strict mode
+# The pattern for working with raw data is this:
+# my $value = do { no strict 'refs'; ... }; ## no critic
+
+# in: none
+# out: sorted & filtered list of symbols
+sub read_names {
+    my $self = shift;
+
+    my $package = $self->{target};
+    my $except  = $self->{except_rex};
+    my $only    = $self->{only_rex};
+
+    my @list = sort grep {
+        /^\w+$/ and $_ !~ $except and $_ =~ $only
+    } do {
+        no strict 'refs'; ## no critic
+        keys %{ $package."::" };
+    };
+
+    return @list;
+};
+
+# In: symbol list arrayref (read_symbols if none)
+# Out: symbol table hashref
 sub read_symbols {
     my ($self, $list) = @_;
 
     my $package = $self->{target};
     $list ||= [ $self->read_names ];
+    $list = [ grep { $self->{restrict_symbols}{$_} } @$list ]
+        if $self->{restrict_symbols};
 
     my %content;
     foreach my $name ( @$list ) {
@@ -379,36 +633,22 @@ sub read_symbols {
     return \%content;
 };
 
-# This method's signature is a bit counterintuitive:
-# $self->replace_symbols( \@names_to_erase, \%new_table_entries )
-# If first argument is omitted, the whole namespace is scanned instead.
-# Separate erase_symbols and write_symbols turned out to be cumbersome
-#    because of the need to handle granular exclusion list.
-# This method can fill both roles.
-# Providing an empty list would make it just write the symbol table,
-#    whereas an empty hash would mean deletion only.
-sub replace_symbols {
-    my ($self, $clear_list, $table) = @_;
+# writes raw symbols, ignoring touch_not!
+# In: symbol table hashref
+# Out: none
+sub write_symbols {
+    my ($self, $table) = @_;
 
     my $package = $self->{target};
 
-    $clear_list ||= [ $self->read_names ];
-    $table ||= {};
+    if (DEBUG) {
+        my $old_table = $self->read_symbols;
+        $self->message( "package $self->{target} to be altered: ".dump_table($table, $old_table) )
+    };
 
-    my %uniq;
-    $uniq{$_}++ for keys %$table, @$clear_list;
 
-    foreach my $name( keys %uniq ) {
-        my $copy = $table->{$name} || {};
-        if ( my $skip = $self->{touch_not}{$name} ) {
-            foreach my $type (keys %$skip) {
-                my $value = do {
-                    no strict 'refs'; ## no critic
-                    *{$package."::".$name}{$type};
-                };
-                $copy->{$type} = $value if defined $value;
-            };
-        };
+    foreach my $name( keys %$table ) {
+        my $copy = $table->{$name};
 
         {
             no strict 'refs'; ## no critic
@@ -432,6 +672,54 @@ sub replace_symbols {
     };
 };
 
+### Logging
+
+sub dump_table {
+    my ($table, $old_table) = @_;
+
+    my @out;
+    foreach my $name( sort keys %$table ) {
+        my $glob = $table->{$name};
+        foreach my $type( sort keys %$glob ) {
+            push @out, "*$name\{$type\}=".(
+                $old_table
+                    ? _is_and_was( $glob->{$type}, $old_table->{$name}{$type} )
+                    : _ref2str( $glob->{$type} )
+            );
+        };
+    };
+
+    return join ", ", @out;
+};
+
+sub _is_and_was {
+    my ($new, $old) = @_;
+
+    if ((refaddr $new || 0) != (refaddr $old || 0)) {
+        return _ref2str( $new )."[was: "._ref2str( $old )."]";
+    } else {
+        return _ref2str( $new )."[unchanged]";
+    };
+};
+
+# TODO find existing?
+sub _ref2str {
+    my $ref = shift;
+
+    return ref $ref
+        ? blessed $ref
+            ? sprintf "%s=%s(0x%x)", ref $ref, reftype $ref, refaddr $ref
+            : sprintf "%s(0x%x)", ref $ref, refaddr $ref
+        : 'undef';
+};
+
+sub message {
+    my ($self, $msg) = @_;
+
+    $msg =~ s/\n$//s;
+    carp "$msg via namespace::local from $self->{origin}";
+};
+
 sub _croak {
     croak ("namespace::local: ".shift);
 };
@@ -490,10 +778,15 @@ L<http://search.cpan.org/dist/namespace-local/>
 
 =head1 SEE ALSO
 
-L<namespace::clean>, L<namespace::sweep>, L<namespace::autoclean>...
+L<namespace::clean> gave the inspiration for this module.
+
+L<namespace::sweep>, L<namespace::autoclean> and probably more also clean
+caller's namespace, but differently.
 
 L<B::Hooks::EndOfScope> is used as a backend.
 
+L<perlmod/Symbol Tables> explains how reading/writing the namespace works.
+
 =head1 LICENSE AND COPYRIGHT
 
 Copyright 2018 Konstantin S. Uvarin, C<< <khedin at gmail.com> >>
